package com.tzld.piaoquan.recommend.server.service;

import com.alibaba.fastjson.JSONObject;
import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.tzld.piaoquan.recommend.server.gen.common.Result;
import com.tzld.piaoquan.recommend.server.gen.recommend.HomepageRecommendRequest;
import com.tzld.piaoquan.recommend.server.gen.recommend.HomepageRecommendResponse;
import com.tzld.piaoquan.recommend.server.gen.recommend.Video;
import com.tzld.piaoquan.recommend.server.model.HomepageRecommendParam;
import com.tzld.piaoquan.recommend.server.service.rank.RankParam;
import com.tzld.piaoquan.recommend.server.service.rank.RankResult;
import com.tzld.piaoquan.recommend.server.service.rank.RankService;
import com.tzld.piaoquan.recommend.server.service.recall.RecallParam;
import com.tzld.piaoquan.recommend.server.service.recall.RecallResult;
import com.tzld.piaoquan.recommend.server.service.recall.RecallService;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.Region24HRecallStrategy;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.RegionHRecallStrategy;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.RegionRelative24HDupRecallStrategy;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.RegionRelative24HRecallStrategy;
import com.tzld.piaoquan.recommend.server.util.DateUtils;
import com.tzld.piaoquan.recommend.server.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.RandomUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.ZSetOperations;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import javax.annotation.PostConstruct;
import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * @author dyp
 */
@Service
@Slf4j
public class RecommendService {

    @Autowired
    private RedisTemplate<String, String> redisTemplate;


    private Map<Integer, String> ab_initial_config_map = new HashMap<>();

    @ApolloJsonValue("ab_exp_code")
    private Map<String, Map<String, String>> abExpCodeMap = new HashMap<>();

    @ApolloJsonValue("app_ab_code")
    private Map<String, Map<String, Map<String, String>>> appAbCodeMap = new HashMap<>();

    private int[] flowPoolIdList = {7, 8, 9, 10, 11, 12, 13, 14, 15, 16};

    @Autowired
    private FlowPoolConfigService flowPoolConfigService;

    @Autowired
    private RecallService recallService;
    @Autowired
    private RankService rankService;
    @Autowired
    private PreViewedService preViewedService;

    @PostConstruct
    public void init() {
        ab_initial_config_map.put(0, "095");
        ab_initial_config_map.put(4, "144");
        ab_initial_config_map.put(5, "121");
        ab_initial_config_map.put(18, "074");
        ab_initial_config_map.put(19, "069");
        ab_initial_config_map.put(-1, "095-1");
    }

    public HomepageRecommendResponse homepageRecommend(HomepageRecommendRequest request) {

        if (request.getVersionAuditStatus() == 1) {
            return specialMidRecommend(request);
        }
        if (StringUtils.isNotBlank(request.getMid())
                && redisTemplate.opsForSet().isMember("special:mid", request.getMid())) {
            return specialMidRecommend(request);
        }

        HomepageRecommendParam param = genHomepageRecommendParam(request, 0);
        log.info("genHomepageRecommendParam : {}", JSONUtils.toJson(param));


        List<Video> videos = videoRecommend(request, param);

        updateCache(request, videos);

        return HomepageRecommendResponse.newBuilder()
                .setResult(Result.newBuilder().setCode(1).setMessage("success"))
                .addAllVideo(videos)
                .build();

    }

    private void updateCache(HomepageRecommendRequest request, List<Video> videos) {
        if (StringUtils.isNotBlank(request.getMid())
                || CollectionUtils.isEmpty(videos)) {
            return;
        }
        // 1 更新预曝光数据 过期时间 30min
        Long[] videoIds = new Long[videos.size()];
        for (int i = 0; i < videoIds.length; i++) {
            videoIds[i] = videos.get(i).getVideoId();
        }

        preViewedService.updateCache(request.getAppType(), request.getMid(), videoIds);
        updateLastVideoCache(request, videos);

        // 更新流量池：本地分发数量-1
        // redis没有这个key 也没有flowPool这个push_form
        // updateFlowPoolCache(request, videos);

        // 更新视频限流逻辑：redis没有这个key，而且代码中无引用

    }

    private void updateLastVideoCache(HomepageRecommendRequest request, List<Video> videos) {
        // 2 地域小时最后一个视频
        for (int i = videos.size() - 1; i >= 0; i--) {
            if (videos.get(i).equals(RegionHRecallStrategy.PUSH_FORM)) {
                redisTemplate.opsForValue().set(String.format(RegionHRecallStrategy.LAST_VIDEO_KEY_FORMAT,
                                request.getAppType(), request.getMid()), String.valueOf(videos.get(i).getVideoId()),
                        24, TimeUnit.HOURS);
            }
        }

        // 3 地域24小时最后一个视频
        for (int i = videos.size() - 1; i >= 0; i--) {
            if (videos.get(i).equals(Region24HRecallStrategy.PUSH_FORM)) {
                redisTemplate.opsForValue().set(String.format(Region24HRecallStrategy.LAST_VIDEO_KEY_FORMAT,
                                request.getAppType(), request.getMid()), String.valueOf(videos.get(i).getVideoId()),
                        24, TimeUnit.HOURS);
            }
        }

        // 4 地域相对24小时最后一个视频
        for (int i = videos.size() - 1; i >= 0; i--) {
            if (videos.get(i).equals(RegionRelative24HRecallStrategy.PUSH_FORM)) {
                redisTemplate.opsForValue().set(String.format(RegionRelative24HRecallStrategy.LAST_VIDEO_KEY_FORMAT,
                                request.getAppType(), request.getMid()), String.valueOf(videos.get(i).getVideoId()),
                        24, TimeUnit.HOURS);
            }
        }

        // 5 地域相对24小时最后一个视频
        for (int i = videos.size() - 1; i >= 0; i--) {
            if (videos.get(i).equals(RegionRelative24HDupRecallStrategy.PUSH_FORM)) {
                redisTemplate.opsForValue().set(String.format(RegionRelative24HDupRecallStrategy.LAST_VIDEO_KEY_FORMAT,
                                request.getAppType(), request.getMid()), String.valueOf(videos.get(i).getVideoId()),
                        24, TimeUnit.HOURS);
            }
        }
    }

    private List<Video> videoRecommend(HomepageRecommendRequest request, HomepageRecommendParam param) {
        // TODO
        RecallParam recallParam = new RecallParam();
        RecallResult recallResult = recallService.recall(recallParam);

        RankParam rankParam = new RankParam();
        rankParam.setRecallResult(recallResult);
        RankResult result = rankService.rank(rankParam);

        if (result == null || CollectionUtils.isEmpty(result.getVideos())) {
            return Collections.emptyList();
        }

        // 只返回size条数据
        List<Video> videos = new ArrayList<>();
        int size = request.getSize() < result.getVideos().size()
                ? request.getSize()
                : result.getVideos().size();

        for (int i = 0; i < size; i++) {
            videos.add(Video.newBuilder()
                    .setPosition(i + 1)
                    .setPushFrom(videos.get(i).getPushFrom())
                    .setAbCode(videos.get(i).getAbCode())
                    .setVideoId(videos.get(i).getVideoId())
                    .setRovScore(videos.get(i).getRovScore())
                    .setSortScore(videos.get(i).getRovScore())
                    .build());
        }

        return videos;
    }

    private HomepageRecommendParam genHomepageRecommendParam(HomepageRecommendRequest request, int recommendType) {
        HomepageRecommendParam param = new HomepageRecommendParam();
        param.setTop_K(3);
        param.setFlowPoolP(0.3);
        param.setNo_op_flag(true);
        param.setExpire_time(3600);
        param.setOld_video_index(-1);

        String abInitialConfig = ab_initial_config_map.containsKey(request.getAppType())
                ? ab_initial_config_map.get(request.getAppType())
                : ab_initial_config_map.get(-1);

        Map<String, String> abExpCode = abExpCodeMap.get(abInitialConfig);
        param.setAb_code(abExpCode.get("ab_code"));
        param.setRule_key(abExpCode.get("rule_key"));
        param.setData_key(abExpCode.get("data_key"));
        param.setRule_key_30day(abExpCode.get("30day_rule_key"));
        param.setShield_config("");
        param.setRank_key_prefix("rank:score1:");

        // 实验配置
        if (StringUtils.isNotBlank(request.getAbExpInfo())) {
            Map<String, List<Map<String, String>>> map = JSONObject.parseObject(request.getAbExpInfo(), Map.class);
            Set<String> abExpCodes = new HashSet<>();
            for (List<Map<String, String>> list : map.values()) {
                list.stream().forEach(m -> abExpCodes.add(m.get("abExpCode")));
            }

            // 流量池视频分发概率
            if (abExpCodes.contains("211")) {
                param.setFlowPoolP(0.9);
            } else if (abExpCodes.contains("221")) {
                param.setFlowPoolP(0.7);
            } else if (abExpCodes.contains("299")) {
                param.setFlowPoolP(0.5);
            } else if (abExpCodes.contains("300")) {
                param.setFlowPoolP(0.4);
            } else if (abExpCodes.contains("301")) {
                param.setFlowPoolP(0.6);
            } else if (abExpCodes.contains("339")) {
                param.setFlowPoolP(0);
            }

            for (Map.Entry<String, Map<String, String>> entry : abExpCodeMap.entrySet()) {
                if (abExpCodes.contains(entry.getKey())) {
                    param.setAb_code(entry.getValue().get("ab_code"));
                    param.setRule_key(entry.getValue().get("rule_key"));
                    param.setData_key(entry.getValue().get("data_key"));
                    param.setRank_key_prefix(StringUtils.isNotBlank(entry.getValue().get("rank_key_prefix"))
                            ? entry.getValue().get("rank_key_prefix")
                            : "rank:score1:");
                    param.setRule_key_30day(entry.getValue().get("30day_rule_key"));
                    param.setShield_config(entry.getValue().get("ab_code"));
                    break;
                }
            }
        }

        // app实验组
        // {"10018":{"eventId":"100182","ifAddedRootEventId":true},"10020":{"eventId":"100201","ifAddedRootEventId":true}}
        if (StringUtils.isNotBlank(request.getAbInfoData())) {
            Map<String, Map<String, String>> map = JSONObject.parseObject(request.getAbInfoData(), Map.class);
            Map<String, String> abInfoAppMap = new HashMap<>();

            for (Map.Entry<String, Map<String, String>> entry : map.entrySet()) {
                if (MapUtils.isEmpty(entry.getValue())) {
                    continue;
                }
                String eventId = entry.getValue().get("eventId");
                if (StringUtils.isNotBlank(eventId)) {
                    abInfoAppMap.put(entry.getKey(), eventId);
                }
            }

            if (recommendType == 0) {
                String eventId = abInfoAppMap.get("10003");
                Map<String, Map<String, String>> appAbCode = appAbCodeMap.get("10003");
                if (StringUtils.isNotBlank(eventId) && appAbCode.containsKey(eventId)) {
                    param.setAb_code(appAbCode.get(eventId).get("ab_code"));
                    param.setRule_key(appAbCode.get(eventId).get("rule_key"));
                    param.setData_key(appAbCode.get(eventId).get("data_key"));
                }
            }

            // 流量池分发实验组划分
            int flowPoolIdChoice = flowPoolIdList[RandomUtils.nextInt(0, flowPoolIdList.length - 1)];
            param.setFlow_pool_abtest_group("control_group");
            try {
                Map<String, List<Integer>> flowPoolConfig = flowPoolConfigService.getFlowPoolConfig();
                for (Map.Entry<String, List<Integer>> entry : flowPoolConfig.entrySet()) {
                    if (entry.getValue().contains(flowPoolIdChoice)) {
                        param.setFlow_pool_abtest_group(entry.getKey());
                    }
                }
            } catch (Exception e) {
                log.error("flowPoolConfigCache get error", e);
            }
        }

        return param;
    }

    private HomepageRecommendResponse specialMidRecommend(HomepageRecommendRequest request) {
        String keyNamePrefix = "special:videos:item:";
        String dateStr = DateUtils.getCurrentDateStr("yyyyMMdd");
        String specialKeyName = keyNamePrefix + dateStr;
        if (!redisTemplate.hasKey(specialKeyName)) {
            dateStr = DateUtils.getBeforeDaysDateStr("yyyyMMdd", 1);
            specialKeyName = keyNamePrefix + dateStr;
        }

        String lastSpecialRecallKey = String.format("recall:last:special:%s:%s:%s", request.getAppType(), request.getMid(), dateStr);
        String value = redisTemplate.opsForValue().get(lastSpecialRecallKey);
        Long idx = 0L;
        if (StringUtils.isNotBlank(value)) {
            idx = redisTemplate.opsForZSet().reverseRank(specialKeyName, value);
            idx = idx == null ? 0L : idx;
        }

        int getSize = request.getSize() * 5;
        int freq = 0;
        List<Video> results = new ArrayList<>();
        while (results.size() < request.getSize()) {
            freq += 1;
            if (freq > 2) {
                break;
            }
            Set<ZSetOperations.TypedTuple<String>> data = redisTemplate.opsForZSet().reverseRangeWithScores(specialKeyName, idx, idx + getSize - 1);
            if (CollectionUtils.isEmpty(data)) {
                break;
            }
            idx += getSize;
            data.stream().forEach(t ->
                    results.add(Video.newBuilder()
                            .setVideoId(Long.getLong(t.getValue(), 0L))
                            .setRovScore(t.getScore())
                            .setAbCode("99999")
                            .setPushFrom("special_mid_videos")
                            .build())
            );
        }

        if (StringUtils.isNotBlank(request.getMid()) && !CollectionUtils.isEmpty(results)) {
            redisTemplate.opsForValue().set(lastSpecialRecallKey, String.valueOf(results.get(results.size() - 1).getVideoId()), 1, TimeUnit.DAYS);
        }

        return HomepageRecommendResponse.newBuilder()
                .setResult(Result.newBuilder().setCode(1).setMessage("success"))
                .addAllVideo(results)
                .build();
    }

    public static void main(String[] args) {
        String json = "{\"ab_test001\":[{\"ab_code\":\"1\",\"config_value\":\"v1\"},{\"ab_code\":\"2\",\"config_value\":\"v2\"}],\"ab_test002\":[{\"ab_code\":\"3\",\"config_value\":\"v3\"},{\"ab_code\":\"4\",\"config_value\":\"v4\"}]}";
        Map<String, List<Map<String, String>>> map = JSONObject.parseObject(json, Map.class);
        System.out.println("");
    }

}
