package com.tzld.piaoquan.recommend.server.service;

import com.alibaba.fastjson.JSONObject;
import com.google.common.collect.Lists;
import com.tzld.piaoquan.recommend.server.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.HttpEntity;
import org.apache.http.NameValuePair;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.message.BasicNameValuePair;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.io.InputStream;
import java.util.*;

/**
 * @author dyp
 */
@Service
@Slf4j
public class ViewedService {
    private Map<Integer, List<Integer>> viewedTypesMap = new HashMap<>();
    private List<Integer> defaultViewedTypes = Lists.newArrayList(1, 6, 7);

    @Value("video.filter.url")
    private String videoFilterUrl;

    private CloseableHttpClient clients = HttpClients.createDefault();

    @PostConstruct
    public void init() {
        viewedTypesMap.put(0, Lists.newArrayList(1, 6));
        viewedTypesMap.put(4, Lists.newArrayList(1, 6));
        viewedTypesMap.put(13, Lists.newArrayList(1));
    }

    public List<Long> filterViewedVideo(int appType, String mid, String uid, List<Long> videoIds) {
        List<Integer> viewedTypes = viewedTypesMap.getOrDefault(appType, defaultViewedTypes);

        CloseableHttpResponse chr = null;

        try {

            HttpPost post = new HttpPost(videoFilterUrl);
            post.addHeader("Connection", "close");

            List<NameValuePair> nvps = new ArrayList<NameValuePair>();
            nvps.add(new BasicNameValuePair("appType", String.valueOf(appType)));
            nvps.add(new BasicNameValuePair("mid", mid));
            nvps.add(new BasicNameValuePair("uid", uid));
            nvps.add(new BasicNameValuePair("types", JSONUtils.toJson(viewedTypes)));
            nvps.add(new BasicNameValuePair("videoIds", JSONUtils.toJson(videoIds)));

            post.setEntity(new UrlEncodedFormEntity(nvps, "UTF-8"));

            chr = clients.execute(post);
            if (chr == null
                    || chr.getStatusLine() == null
                    || chr.getStatusLine().getStatusCode() != 200) {
                log.info("request videoFilterUrl exception! videoFilterUrl={},statusCode={}", videoFilterUrl,
                        chr != null && chr.getStatusLine() != null
                                ? chr.getStatusLine().getStatusCode()
                                : -1);
                return Collections.emptyList();
            }
            HttpEntity entity = chr.getEntity();
            if (entity == null) {
                return Collections.emptyList();
            }
            InputStream is = entity.getContent();
            if (is == null) {
                return Collections.emptyList();
            }
            Map<String, List<Long>> data = JSONObject.parseObject(is.toString(), Map.class);
            if (data.get("code") == null
                    || !data.get("code").equals("0")) {
                return Collections.emptyList();
            }
            return data.get("data");

        } catch (Exception e) {
            log.error("invoke http filterViewedVideo error", e);
        } finally {
            try {
                if (chr != null) {
                    chr.close();
                }
            } catch (Exception ex) {
                log.error("close CloseableHttpResponse error", ex);
            }
        }
        return Collections.emptyList();
    }

}
