package com.tzld.piaoquan.recommend.server.service.rank;

import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.*;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.RandomUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;

import java.util.*;
import java.util.stream.Collectors;

/**
 * @author dyp
 */
public class RankService {
    @Autowired
    private RedisTemplate<String, String> redisTemplate;

    public RankResult rank(RankParam param) {

        if (param.getRecallResult() == null
                || (CollectionUtils.isEmpty(param.getRecallResult().getRovPoolRecall())
                && CollectionUtils.isEmpty(param.getRecallResult().getFlowPoolRecall()))) {
            return null;
        }

        // rank rov recall
        List<Video> rov_recall_rank = rovRecallRank(param);

        // rank flow pool recall
        List<Video> flow_recall_rank = param.getRecallResult().getFlowPoolRecall().stream()
                .collect(Collectors.toList());
        Collections.sort(flow_recall_rank, Comparator.comparingDouble(o -> -o.getRovScore()));


        // TODO 重构
        //    去重原则：
        //        如果视频在ROV召回池topK，则保留ROV召回池，否则保留流量池
        Set<Long> rovTopKVideoIds = new HashSet<>();
        for (int i = 0; i < param.getTopK() && i < rov_recall_rank.size(); i++) {
            rovTopKVideoIds.add(rov_recall_rank.get(i).getVideoId());
        }

        Set<Long> flowPoolVideoIds = new HashSet<>();
        Iterator<Video> flowRecallRankIte = flow_recall_rank.iterator();
        while (flowRecallRankIte.hasNext()) {
            Video data = flowRecallRankIte.next();
            if (rovTopKVideoIds.contains(data.getVideoId())) {
                flowRecallRankIte.remove();
            } else {
                flowPoolVideoIds.add(data.getVideoId());
            }
        }

        Iterator<Video> rovRecallRankIte = rov_recall_rank.iterator();
        while (rovRecallRankIte.hasNext()) {
            Video data = rovRecallRankIte.next();
            if (flowPoolVideoIds.contains(data.getVideoId())) {
                rovRecallRankIte.remove();
            }
        }

        // 融合排序
        if (CollectionUtils.isEmpty(rov_recall_rank)) {
            if (param.getSize() < flow_recall_rank.size()) {
                return new RankResult(flow_recall_rank.subList(0, param.getSize()));
            } else {
                return new RankResult(flow_recall_rank);
            }
        }

        List<Video> datas = new ArrayList<>();
        for (int i = 0; i < param.getTopK() && i < rov_recall_rank.size(); i++) {
            datas.add(rov_recall_rank.get(i));
        }

        double flowPoolP = param.getFlowPoolP();
        if (param.getRecallResult().isQuickPool()) {
            String quick_flow_pool_P = redisTemplate.opsForValue().get("flow:pool:quick:distribute:rate:3");
            flowPoolP = NumberUtils.toDouble(quick_flow_pool_P, 0);
        }

        int flowPoolIndex = 0;
        int rovPoolIndex = param.getTopK() - 1;

        for (int i = 0; i < param.getSize() - param.getTopK(); i++) {
            double rand = RandomUtils.nextDouble(0, 1);
            if (rand < flowPoolP) {
                if (flowPoolIndex < flow_recall_rank.size()) {
                    datas.add(flow_recall_rank.get(flowPoolIndex++));
                } else {
                    break;
                }
            } else {
                if (rovPoolIndex < rov_recall_rank.size()) {
                    datas.add(rov_recall_rank.get(rovPoolIndex++));
                } else {
                    break;
                }
            }
        }
        if (rovPoolIndex >= rov_recall_rank.size()) {
            for (int i = flowPoolIndex; i < flow_recall_rank.size(); i++) {
                datas.add(flow_recall_rank.get(i));
            }
        }
        if (flowPoolIndex >= flow_recall_rank.size()) {
            for (int i = rovPoolIndex; i < rov_recall_rank.size(); i++) {
                datas.add(rov_recall_rank.get(i));
            }
        }
        return new RankResult(datas);
    }

    private List<Video> rovRecallRank(RankParam param) {
        if (param.getAbCode().equals("60054")
                || param.getAbCode().equals("60068")
                || param.getAbCode().equals("60081")
                || param.getAbCode().equals("60084")) {

            List<String> videoIdKeys = param.getRecallResult().getRovPoolRecall().stream()
                    .map(t -> param.getRankKeyPrefix() + t.getVideoId())
                    .collect(Collectors.toList());
            List<Video> rov_recall_rank = param.getRecallResult().getRovPoolRecall().stream()
                    .collect(Collectors.toList());
            List<String> video_scores = redisTemplate.opsForValue().multiGet(videoIdKeys);
            if (CollectionUtils.isNotEmpty(video_scores)
                    && video_scores.size() == rov_recall_rank.size()) {
                for (int i = 0; i < video_scores.size(); i++) {
                    rov_recall_rank.get(i).setRovScore(NumberUtils.toDouble(video_scores.get(i), 0.0));
                }
                Collections.sort(rov_recall_rank, Comparator.comparingDouble(o -> -o.getRovScore()));
            }
            return rov_recall_rank;
        } else {
            List<Video> rov_recall_rank = new ArrayList<>();
            rov_recall_rank.addAll(extractRovPoolAndSort(param, RegionHRecallStrategy.PUSH_FORM));
            rov_recall_rank.addAll(extractRovPoolAndSort(param, Region24HRecallStrategy.PUSH_FORM));
            rov_recall_rank.addAll(extractRovPoolAndSort(param, RegionRelative24HRecallStrategy.PUSH_FORM));
            rov_recall_rank.addAll(extractRovPoolAndSort(param, RegionRelative24HDupRecallStrategy.PUSH_FORM));
            rov_recall_rank.addAll(extractRovPoolAndSort(param, RegionRelative48HRecallStrategy.PUSH_FORM));
            rov_recall_rank.addAll(extractRovPoolAndSort(param, RegionRelative48HDupRecallStrategy.PUSH_FORM));
            return rov_recall_rank;
        }
    }

    private List<Video> extractRovPoolAndSort(RankParam param, String pushForm) {
        if (param == null
                || param.getRecallResult() == null
                || CollectionUtils.isEmpty(param.getRecallResult().getRovPoolRecall())) {
            return Collections.emptyList();
        }
        return param.getRecallResult().getRovPoolRecall().stream()
                .filter(o -> o.getPushFrom().equals(pushForm))
                .sorted(Comparator.comparingDouble(o -> -o.getRovScore()))
                .collect(Collectors.toList());
    }

}
